Option Explicit On
Option Strict On

'Define the DebuggerVisualizer attributes for all of the different classes we are able to visualise.
'All of these point to this class to implement the visualizer.
'Collections classes
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Collections.ArrayList), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Collections.BitArray), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Collections.Hashtable), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Collections.Queue), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Collections.SortedList), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Collections.Stack), Description:="Collection Visualizer")> 
'Collections.Specialized classes
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Specialized.HybridDictionary), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Specialized.ListDictionary), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Specialized.NameObjectCollectionBase), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Specialized.OrderedDictionary), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Specialized.StringCollection), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(Specialized.StringDictionary), Description:="Collection Visualizer")> 
'Collections.Generic classes
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.Dictionary(Of ,)), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.List(Of )), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.LinkedList(Of )), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.Queue(Of )), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.SortedDictionary(Of ,)), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.SortedList(Of ,)), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(collections.Generic.Stack(Of )), Description:="Collection Visualizer")> 
'Other classes
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(CollectionBase), Description:="Collection Visualizer")> 
<Assembly: DebuggerVisualizer(GetType(CollectionVisualizer), GetType(ObjectSource), Target:=GetType(microsoft.VisualBasic.Collection), Description:="Collection Visualizer")> 



''' <summary>
''' CollectionVisualizer implements the visualizer code
''' </summary>
''' <remarks></remarks>
Public Class CollectionVisualizer
    Inherits Microsoft.VisualStudio.DebuggerVisualizers.DialogDebuggerVisualizer

	''' <summary>
	''' Called by the visualizer framework to display the visualization of the target object.
	''' </summary>
	''' <param name="windowService"></param>
	''' <param name="objectProvider"></param>
	''' <remarks></remarks>
    Protected Overrides Sub Show(ByVal windowService As Microsoft.VisualStudio.DebuggerVisualizers.IDialogVisualizerService, ByVal objectProvider As Microsoft.VisualStudio.DebuggerVisualizers.IVisualizerObjectProvider)

        Dim preview As New PreviewForm
        Dim dt As DataTable

		'Read the collection data we have stored back into a datatable
        dt = DirectCast(objectProvider.GetObject, DataTable)

		'Display the data within the preview form
        preview.ShowData(dt)

	End Sub

End Class

''' <summary>
''' A class to provide the ObjectSource for the visualizer.
''' </summary>
''' <remarks>Some of the collection types are not able to be serialized in order for their
''' data to be passed to the CollectionVisualizer. The ObjectSource class runs prior
''' to serialization, and packages up all the collection data into a DataTable. This
''' DataTable is then passed over to the visualizer in order for it to obtain the
''' data to be displayed.</remarks>
Public Class ObjectSource
    Inherits Microsoft.VisualStudio.DebuggerVisualizers.VisualizerObjectSource

    Public Overrides Sub GetData(ByVal target As Object, ByVal outgoingData As System.IO.Stream)

        Dim dtData As DataTable

		'Which type of collection are we visualizing? Determine the type, then copy all
		'of the relevant data from the collection into the DataTable.
        Select Case True

			Case TypeOf target Is Specialized.NameValueCollection
				dtData = PopulateCollectionInfo(DirectCast(target, Specialized.NameValueCollection))

            Case TypeOf target Is Web.HttpCookieCollection
				dtData = PopulateCollectionInfo(DirectCast(target, Web.HttpCookieCollection))

            Case TypeOf target Is Specialized.NameObjectCollectionBase
				dtData = PopulateCollectionInfo(DirectCast(target, Specialized.NameObjectCollectionBase))

            Case TypeOf target Is Microsoft.VisualBasic.Collection
				dtData = PopulateCollectionInfo(DirectCast(target, Collection))

            Case TypeOf target Is BitArray
				dtData = PopulateCollectionInfo(DirectCast(target, BitArray))

            Case TypeOf target Is Queue
				dtData = PopulateCollectionInfo(DirectCast(target, Queue))

            Case TypeOf target Is Stack
				dtData = PopulateCollectionInfo(DirectCast(target, Stack))


			Case TryCast(target, Collections.IDictionary) IsNot Nothing
				dtData = PopulateCollectionInfo(DirectCast(target, Collections.IDictionary))

            Case TryCast(target, Collections.IList) IsNot Nothing
				dtData = PopulateCollectionInfo(DirectCast(target, Collections.IList))

			Case TryCast(target, Collections.ICollection) IsNot Nothing
				dtData = PopulateCollectionInfo(DirectCast(target, Collections.ICollection))

			Case TryCast(target, Collections.IEnumerable) IsNot Nothing
				dtData = PopulateCollectionInfo(DirectCast(target, Collections.IEnumerable))


			Case Else
				Throw New ApplicationException("Unknown object type: " & TypeName(target))

		End Select

		'Set the tablename to indicate the collection type
        dtData.TableName = TypeName(target)

		'Ensure all data added to the table has been accepted
        dtData.AcceptChanges()

		'Pass the datatable through to the visualizer to display
        Microsoft.VisualStudio.DebuggerVisualizers.VisualizerObjectSource.Serialize(outgoingData, dtData)

	End Sub

	''' <summary>
	''' Populate the DataTable with values from a NameValueCollection or derivative
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
    Private Function PopulateCollectionInfo(ByVal coll As Specialized.NameValueCollection) As DataTable

        Dim dtData As DataTable = InitDataTable(2)
        Dim dr As DataRow
        Dim i As Integer

		'Store the name and value for each item
        For i = 0 To coll.Count - 1
            dr = dtData.NewRow
            dtData.Rows.Add(dr)
            dr("Name") = coll.Keys(i)
            dr("Value") = coll.Item(i)
        Next

        Return dtData

    End Function

	''' <summary>
	''' Populate the DataTable with values from an HttpCookieCollection or derivative
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Web.HttpCookieCollection) As DataTable

		Dim dtData As DataTable = InitDataTable(3)
		Dim dr As DataRow
		Dim i As Integer
		Dim cookie As Web.HttpCookie

		'Store the name, value and expiry date/time for each item.
		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			cookie = coll.Item(i)
			dr("Name") = cookie.Name
			dr("Value") = cookie.Value
			If cookie.Expires > DateTime.MinValue Then
				'Format the date is ISO8601 format
				dr("Expires") = cookie.Expires.ToString("yyyy-MM-dd HH:mm:ss")
			Else
				dr("Expires") = ""
			End If
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from a NameObjectCollectionBase or derivative
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Specialized.NameObjectCollectionBase) As DataTable

		Dim dtData As DataTable = InitDataTable(2)
		Dim dr As DataRow
		Dim i As Integer
		Dim members() As Reflection.MemberInfo
		Dim keys As Array
		Dim keysArray() As Object
		Dim values As Array
		Dim valuesArray() As Object
		Dim valueItem As Object
		Dim thisValue As String

		'Reading from this type of object is more difficult as the BaseGetAllValues and BaseGetAllKeys
		'members of the class aren't public. We'll use a little trickery to gain access to them.

		'Find the BaseGetAllValues member
		members = coll.GetType.FindMembers(Reflection.MemberTypes.Method, Reflection.BindingFlags.Public Or Reflection.BindingFlags.NonPublic Or Reflection.BindingFlags.Instance, Type.FilterName, "BaseGetAllValues")
		'Found it? (Note that there are two versions, the second takes parameters, we want the first)
		If members.Length = 0 Then
			Throw New ApplicationException("Unable to retrieve values from NameObjectCollectionBase")
		End If
		Dim mi As Reflection.MethodInfo = DirectCast(members(0), Reflection.MethodInfo)
		'Get the array of objects that it returns
		values = DirectCast(mi.Invoke(coll, Nothing), Array)
		'Cast to an object array for later use
		valuesArray = CType(values, Object())

		'Repeat for BaseGetAllKeys
		members = coll.GetType.FindMembers(Reflection.MemberTypes.Method, Reflection.BindingFlags.Public Or Reflection.BindingFlags.NonPublic Or Reflection.BindingFlags.Instance, Type.FilterName, "BaseGetAllKeys")
		If members.Length = 0 Then
			Throw New ApplicationException("Unable to retrieve keys from NameObjectCollectionBase")
		End If
		mi = DirectCast(members(0), Reflection.MethodInfo)
		'Invoke it and get the return value. This will be an array of ArrayList objects
		keys = DirectCast(mi.Invoke(coll, Nothing), Array)
		'Cast to an object array for later use
		keysArray = CType(keys, Object())

		'Now we have the keys and the values, we can build up the datatable
		For i = 0 To UBound(keysArray)
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			'Get the key name
			dr("Name") = GetObjectValue(keysArray(i))

			'Build up all values stored against this item into a comma-separated string
			thisValue = ""
			Select Case True
				Case TypeOf valuesArray(i) Is ArrayList
					'Multiple items have been stored with this name, so build them all into a single value
					For Each valueItem In DirectCast(valuesArray(i), ArrayList)
						If Len(thisValue) > 0 Then thisValue &= ","
						thisValue &= GetObjectValue(valueItem)
					Next
				Case TypeOf valuesArray(i) Is System.Web.HttpCookie
					'This is a cookie, so read out something more meaningful
					thisValue = DirectCast(valuesArray(i), System.Web.HttpCookie).Value
				Case Else
					thisValue = GetObjectValue(valuesArray(i))
			End Select

			dr("Value") = thisValue
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from a BitArray or derivative
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As BitArray) As DataTable

		Dim dtData As DataTable = InitDataTable(2)
		Dim dr As DataRow
		Dim i As Integer

		dtData.Columns("Name").Caption = "Index"

		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			dr("Name") = i
			dr("Value") = GetObjectValue(coll(i))
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from a Queue or derivative
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Queue) As DataTable

		Dim dtData As DataTable = InitDataTable(2)
		Dim dr As DataRow
		Dim i As Integer
		Dim values() As Object

		'Instead of the name, we'll write the position within the queue for each item
		dtData.Columns("Name").Caption = "Position"

		'Retrieve a copy of all the queued values as an array to make them easier to access
		values = coll.ToArray()

		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			'Store the position and the value
			dr("Name") = i
			dr("Value") = GetObjectValue(values(i))
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from a Stack or derivative
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Stack) As DataTable

		Dim dtData As DataTable = InitDataTable(2)
		Dim dr As DataRow
		Dim i As Integer
		Dim values() As Object

		'Instead of the name, we'll write the position within the queue for each item
		dtData.Columns("Name").Caption = "Position"

		'Retrieve a copy of all the queued values as an array to make them easier to access
		values = coll.ToArray()

		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			'Store the position and the value
			dr("Name") = i
			dr("Value") = GetObjectValue(values(i))
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from an object that implements the IDictionary interface
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Collections.IDictionary) As DataTable

		Dim dtData As DataTable = InitDataTable(2)
		Dim dr As DataRow
		Dim i As Integer
		Dim keys(coll.Count - 1) As Object
		Dim values(coll.Count - 1) As Object

		coll.Keys.CopyTo(keys, 0)
		coll.Values.CopyTo(values, 0)

		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			dr("Name") = keys(i).ToString
			dr("Value") = values(i).ToString
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from an object that implements the IList interface
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Collections.IList) As DataTable

		Dim dtData As DataTable = InitDataTable(1)
		Dim dr As DataRow
		Dim i As Integer

		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			dr("Value") = GetObjectValue(coll.Item(i))
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from an object that implements the ICollection interface
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Collections.ICollection) As DataTable

		Dim dtData As DataTable = InitDataTable(1)
		Dim dr As DataRow
		Dim i As Integer
		Dim values(coll.Count - 1) As Object

		coll.CopyTo(values, 0)

		For i = 0 To coll.Count - 1
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			dr("Value") = values(i).ToString
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from an object that implements the IEnumerable interface
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks></remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Collections.IEnumerable) As DataTable

		Dim dtData As DataTable = InitDataTable(2)
		Dim dr As DataRow
		Dim d As DictionaryEntry

		For Each d In coll
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			dr("Name") = GetObjectValue(d.Key)
			dr("Value") = GetObjectValue(d.Value)
		Next

		Return dtData

	End Function

	''' <summary>
	''' Populate the DataTable with values from a VB6-style VisualBasic.Collection
	''' </summary>
	''' <param name="coll"></param>
	''' <returns></returns>
	''' <remarks>Sadly we are unable to retrieve the key values from this type of colleciton, as they
	''' are stored as a hash rather than as their actual values.</remarks>
	Private Function PopulateCollectionInfo(ByVal coll As Microsoft.VisualBasic.Collection) As DataTable

		Dim dtData As DataTable = InitDataTable(1)
		Dim dr As DataRow
		Dim o As Object

		For Each o In coll
			dr = dtData.NewRow
			dtData.Rows.Add(dr)
			dr("Value") = GetObjectValue(o)
		Next

		Return dtData

	End Function

	''' <summary>
	''' Get a string representation of the supplied object.
	''' </summary>
	''' <param name="Obj"></param>
	''' <returns>If the object is Nothing, returns Nothing. Otherwise calls .ToString() on the object and returns the obtained value</returns>
	''' <remarks></remarks>
	Private Function GetObjectValue(ByVal Obj As Object) As String

		If Obj Is Nothing Then
			Return Nothing
		Else
			Return Obj.ToString
		End If

	End Function

	''' <summary>
	''' Prepare the DataTable to store the values read from our collection.
	''' </summary>
	''' <param name="Columns">The number of columns to include within the datatable.
	''' A value of 1 will create just a Value column.
	''' 2 will also create a Name column.
	''' 3 will also create an Expires column</param>
	''' <returns></returns>
	''' <remarks></remarks>
    Private Function InitDataTable(ByVal Columns As Integer) As DataTable

        Dim dtData As New DataTable

        'Set the scheme columns
		With dtData
			'If we have 2 or more columns then we need to include the Name column
			If Columns >= 2 Then
				.Columns.Add("Name")
			End If

			'We always include the Value column
			.Columns.Add("Value")

			'If we have 3 or more columns then we need ot include the Expires column
			If Columns >= 3 Then
				'Add the Expires column. We'll always store dates in textual ISO8601-style format for consistency
				.Columns.Add("Expires")
			End If

		End With

        Return dtData

    End Function

End Class